@@ -9,6 +9,7 @@ from attrdict import AttrDict
 import attrdict
 from getpass import getpass
 from queue import Queue
+from threading import *
 
 class Bot:
 
@@ -64,9 +65,9 @@ class Bot:
         self._copy = False
         for item in vars:
             if item in kwargs:
-                self.setattr(item, kwargs[item])
+                setattr(self, item, kwargs[item])
             elif item in noneVars:
-                self.setattr(item, None)
+                setattr(self, item, None)
         self.normowner = re.sub(r"\s+", "", self.owner)
         self.parser = argparse.ArgumentParser(description=f'{nick}: A euphoria.io bot.' if 'description' not in kwargs else kwargs['description'])
         if 'customArgs' in kwargs and kwargs['customArgs']:
@@ -77,10 +78,10 @@ class Bot:
             self.add_command_line_arg("--room", "-r", help = f"Set the room the bot will be placed in. Default: {'bots' if 'room' not in kwargs else kwargs['room']}", default = "bots" if "room" not in kwargs else kwargs["room"])
             self.add_command_line_arg("--password", "-p", help = "Set the password for the room the bot will be placed in.", default = "" if "password" not in kwargs else kwargs["password"])
             self.args = self.parser.parse_args()
-            self.room = self.args.room if self.args.test != True or _copy else "test"
+            self.room = self.args.room if self.args.test != True or self._copy else "test"
             self.password = self.args.password
             if not self._copy:
-                log(f"Debug: {str(self.args.test)}", _copy = self._copy)
+                Bot.log(f"Debug: {str(self.args.test)}", _copy = self._copy)
 
     def _wait_for(self, response):
         reply = None
@@ -95,7 +96,7 @@ class Bot:
             else:
                 waitlist.put(msg)
             if i > self.APITimeout:
-                log(f"{response} API response not recorded. reason: too many events before {response}.", error = True, _copy = self._copy)
+                Bot.log(f"{response} API response not recorded. reason: too many events before {response}.", error = True, _copy = self._copy)
                 reply = None
                 break
         while not waitlist.empty():
@@ -113,15 +114,16 @@ class Bot:
                 self.kill()
             if re.search(f'^!restart @{re.escape(self.normname)}$', msg.data.content) is not None:
                 self.restart(msg = msg)
+                raise KilledError(f"old {self.nick} in &{self.room} killed. Restarted.")
         if msg.type == 'error':
-            log(msg, error = True, _copy = self._copy)
+            Bot.log(msg, error = True, _copy = self._copy)
         elif msg.type in self.handlers:
             self.handlers[msg.type](msg)
         else:
             if callable(self.function):
                 self.function(msg)
             else:
-                log("Advanced start must be given a callable message handler function that takes an AttrDict as its argument.", error = True, _copy = self._copy)
+                Bot.log("Advanced start must be given a callable message handler function that takes an AttrDict as its argument.", error = True, _copy = self._copy)
 
     def _is_privileged(self, user):
         if "is_manager" in user or user.name == self.owner:
@@ -133,12 +135,6 @@ class Bot:
         self.send_command(type = 'ping-reply', data = {'time': msg.data.time})
 
     def _handle_message(self, msg):
-        if re.search(f'^!kill @{re.escape(self.normname)}$', msg.data.content) is not None:
-            self._handle_kill(msg)
-        if re.search(f'^!forcekill @{re.escape(self.normname)}$', msg.data.content) is not None:
-            self.kill()
-        if re.search(f'^!restart @{re.escape(self.normname)}$', msg.data.content) is not None:
-            self.restart(msg = msg)
         if re.search('^!ping$', msg.data.content) is not None or re.search(f'^!ping @(?i){re.escape(self.normname)}$', msg.data.content) is not None:
             self.send_msg(self.ping, parent = msg)
         if re.search(f'^!help @(?i){re.escape(self.normname)}$', msg.data.content) is not None:
@@ -191,27 +187,27 @@ class Bot:
         reply = self._wait_for("auth-reply")
         if reply is not None:
             if reply.data.success == True:
-                log(f'Successfully logged into {self.room}.', _copy = self._copy)
+                Bot.log(f'Successfully logged into {self.room}.', _copy = self._copy)
             else:
-                log(f'Login unsuccessful. Reason: {reply.data.reason}', error = True, _copy = self._copy)
+                Bot.log(f'Login unsuccessful. Reason: {reply.data.reason}', error = True, _copy = self._copy)
                 self._handle_auth(getpass("Enter the correct password: "))
 
     def connect(self):
-        self.conn = websocket.create_connection(f'wss://euphoria.io/room/{self.room}/ws{"?h=0" if not self.human else "?h=1"}', cookie = self.cookie, enable_multithread = True)
-        if self.cookie is None:
-            self.cookie = self.conn.headers['set-cookie']
-        hello = self.recv()
-        reply = self.recv()
+        self.conn = websocket.create_connection(f'wss://euphoria.io/room/{self.room}/ws{"?h=0" if not self.human else "?h=1"}', cookie = self._cookie, enable_multithread = True)
+        if self._cookie is None:
+            self._cookie = self.conn.headers['set-cookie']
+        hello = self._recv()
+        reply = self._recv()
         self._handle_ping(reply)
-        reply = self.recv()
+        reply = self._recv()
         if reply.type == "snapshot-event":
             self.set_nick(self.nick)
         elif reply.type == "bounce-event":
             self._handle_auth(self.password)
             self.set_nick(self.nick)
         else:
-            log(reply, error = True, _copy = self._copy)
-        log(f'connected to &{self.room}.', _copy = self._copy)
+            Bot.log(reply, error = True, _copy = self._copy)
+        Bot.log(f'connected to &{self.room}.', _copy = self._copy)
         return hello
 
     def send_msg(self, msgString, parent = None):
@@ -226,34 +222,37 @@ class Bot:
                 raise BadMessageError(f'message not sent, because type of parent was {type(parent)}. \nParent printed: \n{parent}')
             reply = self._wait_for("send-reply")
             if isinstance(parent, AttrDict):
-                log(f'Message sent: {reply} replying to: {parent.data.id} by {parent.data.sender.name}', _copy = self._copy)
+                Bot.log(f'Message sent: {reply} replying to: {parent.data.id} by {parent.data.sender.name}', _copy = self._copy)
             elif isinstance(parent, str):
-                log(f'Message sent: {reply} replying to: {parent}', _copy = self._copy)
+                Bot.log(f'Message sent: {reply} replying to: {parent}', _copy = self._copy)
             elif parent is None:
-                log(f'Message sent: {reply}', _copy = self._copy)
+                Bot.log(f'Message sent: {reply}', _copy = self._copy)
         except BadMessageError:
             pass
 
-    def restart(self, msg = None, _copyStart = False):
+    def restart(self, msg = None, advanced = False):
         if self.conn is not None:
-            self.conn.close()
+            self.conn.close(timeout = 0)
             del self.conn
-            log("restarting...", _copy = self._copy)
+            Bot.log("restarting...", _copy = self._copy)
         else:
-            log("starting...", _copy = self._copy)
+            Bot.log("starting...", _copy = self._copy)
         self.connect()
-        self.start()
+        t = Thread(target = self.start, kwargs = {"advanced" : advanced})
+        t.start()
+        t.join(0)
 
     def start(self, function = None, advanced = False):
         if not advanced and not self.advanced:
             self.set_handler('ping-event', self._handle_ping)
             self.set_handler('send-event', self._handle_message)
             self.set_handler('bounce-event', self._handle_auth)
+            self.function = lambda msg: None
         else:
             if self.function is None and function is not None and callable(function):
                 self.function = function
             else:
-                log("No valid function passed to advanced bot's start function. Using filler function", _copy = self._copy)
+                Bot.log("No valid function passed to advanced bot's start function. Using filler function", _copy = self._copy)
                 self.function = lambda msg: None
         try:
             for msgJSON in self.conn:
@@ -297,8 +296,10 @@ class Bot:
         self.send_command(type = 'nick', data = {'name': nick})
 
     def kill(self):
-        self.conn.close()
-        raise KilledError(f'{self.normname} killed.')
+        if self.conn is not None and self.conn.connected:
+            self.conn.close()
+        del self.conn
+        raise KilledError(f'{self.normname} in &{self.room} killed.')
 
     def get_userlist(self):
         self.send_command(type = 'who')
@@ -309,15 +310,15 @@ class Bot:
     def move_to(self, roomName, password = "", _copyStart = False):
         self.room = roomName
         self.password = password
-        self.restart(_copyStart = _copyStart)
+        self.restart()
         if _copyStart:
-            log(f"Copy started in &{self.room}.", _copy = self._copy)
+            Bot.log(f"Copy started in &{self.room}.", _copy = self._copy)
 
     def set_handler(self, eventString, function):
         if callable(function):
             self.handlers[eventString] = function
         else:
-            log(f"WARNING: handler for {eventString} not callable, handler not set.", error = True, _copy = self._copy)
+            Bot.log(f"WARNING: handler for {eventString} not callable, handler not set.", error = True, _copy = self._copy)
 
     def set_handlers(self, handlers):
         for eventString, function in handlers.items():
@@ -339,16 +340,16 @@ class Bot:
             function = self.function,
             sendable = self.sendable,
             _copy = True,
-            _cookie = self.cookie,
+            _cookie = self._cookie,
             human = self.human)
         copy.set_handlers(self.handlers)
         copy.set_regexes(self._regexes)
-        t = threading.Thread(target = copy.move_to)
+        t = Thread(target = copy.move_to, kwargs = {"roomName": roomName, "password" : password, "_copyStart" : True})
         self.children.append(copy)
         copy.parents.append(self)
         self.childThreads.append(t)
-        copy.parentThreads.append(threading.current_thread())
-        t.run(roomName = roomName, password = password, _copyStart = True)
+        copy.parentThreads.append(current_thread())
+        t.start()
         t.join(0)
         return copy
 
@@ -370,14 +371,14 @@ class Bot:
             if msg.type == "login-reply":
                 reply = msg
                 if reply.data.success:
-                    log("Login successful.", _copy = self._copy)
+                    Bot.log("Login successful.", _copy = self._copy)
                     self.account = reply.data.account_id
                 else:
-                    log(f"Access denied. Reason: {reply.data.reason}", error = True, _copy = self._copy)
+                    Bot.log(f"Access denied. Reason: {reply.data.reason}", error = True, _copy = self._copy)
                     self.login(email = input("Enter your email: "), password = getpass("Enter your password: "), setDefaults = setDefaults)
                 return reply
             if i > self.APITimeout:
-                log("login-reply API response not recorded. reason: too many events before pm-initiate-reply.", error = True, _copy = self._copy)
+                Bot.log("login-reply API response not recorded. reason: too many events before pm-initiate-reply.", error = True, _copy = self._copy)
                 return None
 
     def initiate_pm(self, id, bot = None, room = "pmtest", hostEmail = None, hostPassword = None):
@@ -396,7 +397,7 @@ class Bot:
                 else:
                     self.copy_to("pm:{to.data.pm_id}")
             else:
-                log("Access Denied: This bot does not have permission to initiate PMs.", error = True, _copy = self._copy)
+                Bot.log("Access Denied: This bot does not have permission to initiate PMs.", error = True, _copy = self._copy)
         return to
 
 class BotError(Exception):
@@ -405,9 +406,9 @@ class BotError(Exception):
 class KilledError(BotError):
     def __init__(self, message = "bot killed."):
         self.message = message
-        log(f'KilledError: {message}', error = True)
+        Bot.log(f'KilledError: {message}', error = True)
 
 class BadMessageError(BotError):
     def __init__(self, message):
         self.message = message
-        log(f'BadMessageError: {message}', error = True)
+        Bot.log(f'BadMessageError: {message}', error = True)
